#pragma once

#include "../../hash/checksum32.h"
#include "../../storage/buffer.h"

#include "../exception.h"

#include <string>
#include <chrono>

#ifdef _WIN32

#include <Windows.h>

#define EVEMON_CLASSNAME    "EVEMON"

namespace debug::logger
{
	struct evemon_header_s
	{
		uint32_t ident;
		uint32_t cksum;
	};

	struct evemon_packet_s
	{
		struct evemon_header_s hdr;

		uint32_t type;
		__time64_t time;

		uint32_t filename_ofs;
		uint32_t message_ofs;
	};

	/* invalid_evemon_event_exception exception class */
	class invalid_evemon_event_exception : public exception::base
	{
	public:
		virtual std::string to_string(void) const override
		{
			return "invalid evemon event object!";
		}
	};

	static const unsigned long EVEMON_VERSION = 'EVE1';

	// pack event
	static storage::buffer pack_event(const event& e)
	{
		storage::buffer ret;

		// allocate header
		ret.allocate<struct evemon_packet_s>();

		// copy data to header
		ret.ptr<struct evemon_packet_s>(0)->hdr.ident = (uint32_t)EVEMON_VERSION;
		ret.ptr<struct evemon_packet_s>(0)->time = std::chrono::system_clock::to_time_t(e.get_time());
		ret.ptr<struct evemon_packet_s>(0)->type = (uint32_t)e.get_type();
		ret.ptr<struct evemon_packet_s>(0)->filename_ofs = (uint32_t)ret.append(e.get_filename());
		ret.ptr<struct evemon_packet_s>(0)->message_ofs = (uint32_t)ret.append(e.get_message());

		// compute cksum
		ret.ptr<struct evemon_packet_s>(0)->hdr.cksum = (uint32_t)checksum32(ret.data() + sizeof(struct evemon_header_s), ret.size() - sizeof(struct evemon_header_s));

		// return element
		return ret;
	}

	// unpack event
	static event unpack_event(const storage::buffer& buffer)
	{
		// check header size
		if (buffer.size() < sizeof(struct evemon_packet_s))
			throw_exception(invalid_evemon_event_exception);

		auto packet = buffer.ptr<struct evemon_packet_s>(0);

		// check ident
		if(packet->hdr.ident != EVEMON_VERSION)
			throw_exception(invalid_evemon_event_exception);

		// check cksum
		if(packet->hdr.cksum != (uint32_t)checksum32(buffer.data() + sizeof(struct evemon_header_s), buffer.size() - sizeof(struct evemon_header_s)))
			throw_exception(invalid_evemon_event_exception);

		// read data
		auto time = std::chrono::system_clock::from_time_t(packet->time);
		auto type = (logger::event::type)packet->type;
		auto filename = buffer.ptr<char>((size_t)packet->filename_ofs);
		auto message = buffer.ptr<char>((size_t)packet->message_ofs);

		// return object;
		return event(time, type, filename, message);
	}

	/* evemon logger class */
	class evemon : public base
	{
	public:

		// ctor
		evemon(void)
		{
			this->m_window_handle = NULL;

			reconnect();
		}

		// cleanup on dtor
		~evemon(void)
		{
			this->m_window_handle = NULL;
		}

		// find window by class name
		void reconnect(void)
		{
			std::lock_guard<std::mutex> lock(this->m_lock);

			this->m_window_handle = FindWindow(TEXT(EVEMON_CLASSNAME), NULL);
		}

		// return true if a connection has been made
		bool is_connected(void) const
		{
			std::lock_guard<std::mutex> lock(this->m_lock);

			return this->m_window_handle != NULL;
		}

		// dispatch events
		virtual void notify(const event& e) const override
		{
			std::lock_guard<std::mutex> lock(this->m_lock);

			// skip if not connected
			if (this->m_window_handle == NULL)
				return;

			// pack event
			auto buffer = pack_event(e);

			// skip message larger than a dword
			if (buffer.size() > ((size_t)1 << (8 * sizeof(DWORD))))
				return;

			// dispatch event
			COPYDATASTRUCT data;

			data.dwData = (int)EVEMON_VERSION;
			data.cbData = (DWORD)buffer.size();
			data.lpData = (void*)buffer.data();

			SendMessage(this->m_window_handle, WM_COPYDATA, (WPARAM)NULL, (LPARAM)&data);
		}

	private:
		mutable std::mutex m_lock;

		HWND m_window_handle;
	};
};

#endif