#pragma once

#include "../../strings/format.h"

#include <string>
#include <regex>
#include <chrono>
#include <list>
#include <ctime>

namespace debug::logger
{
	/* event class */
	class event
	{
	public:
		enum class type
		{
			DBG_MESSAGE = 0,
			DBG_WARNING = 1,
			DBG_ERROR = 2,
			DBG_CRITICAL = 3,
			DBG_EXCEPTION = 4,
			DBG_EXDEBUG = 5,
		};

	public:
		event(std::chrono::system_clock::time_point time, type etype, const std::string& filename, const std::string& message)
		{
			this->m_time = time;
			this->m_type = etype;
			this->m_filename = filename;
			this->m_message = message;
		}

		// get time
		auto get_time(void) const
		{
			return this->m_time;
		}

		// get type
		auto get_type(void) const
		{
			return this->m_type;
		}

		// get filename
		auto get_filename(void) const
		{
			return this->m_filename;
		}

		// get message
		auto get_message(void) const
		{
			return this->m_message;
		}

		// conver time to debug time
		std::string get_string_time(void) const
		{
			struct tm s;
			auto t = std::chrono::system_clock::to_time_t(this->m_time);

#ifdef __GCC__
			localtime_r(&t, &s);
#else
			localtime_s(&s, &t);
#endif

			unsigned int ms = std::chrono::duration_cast<std::chrono::milliseconds>(this->m_time.time_since_epoch()).count() % 1000;

			return format("%.02d:%.02d:%.02d.%.03d", s.tm_hour, s.tm_min, s.tm_sec, ms);
		}

		// convert type to a string
		std::string get_string_type(void) const
		{
			switch (this->m_type)
			{
			case type::DBG_MESSAGE:
				return "message";

			case type::DBG_WARNING:
				return "warning";

			case type::DBG_ERROR:
				return "error";

			case type::DBG_CRITICAL:
				return "critical";

			case type::DBG_EXCEPTION:
				return "exception";

			case type::DBG_EXDEBUG:
				return "extended debug";

			default:
				return "unknown";
			}
		}

		// check for match
		bool match(const std::list<type>& types, const std::list<std::regex>& filenames, const std::list<std::regex>& messages) const
		{
			// filter type if given
			if (types.size() > 0)
			{
				bool success = false;

				for (auto& t : types)
					if (this->m_type == t)
					{
						success = true;

						break;
					}

				if (!success)
					return false;
			}

			// filter filename if any
			if (filenames.size() > 0)
			{
				bool success = false;

				for (auto& t : filenames)
					if (std::regex_match(this->m_filename, t))
					{
						success = true;

						break;
					}

				if (!success)
					return false;
			}

			// filter message if any
			if (messages.size() > 0)
			{
				bool success = false;

				for (auto& t : messages)
					if (std::regex_match(this->m_message, t))
					{
						success = true;

						break;
					}

				if (!success)
					return false;
			}

			// successful
			return true;
		}

	private:
		std::chrono::system_clock::time_point m_time;
		type m_type;
		std::string m_filename, m_message;
	};
};