#pragma once

#include "../../utils/utils.h"
#include "../../utils/singleton.h"
#include "../../threads/lock.h"

#include "../exception.h"

#include "../../exception.h"

#include "base.h"
#include "event.h"

#include <list>
#include <memory>
#include <mutex>
#include <regex>
#include <chrono>

// force extended debugging in _DEBUG mode
#ifdef _DEBUG
#define _EXDEBUG
#endif

namespace debug::logger
{
    /* manager singleton class */
    class manager : public singleton<manager>
    {
        friend class singleton<manager>;

    public:

        // notify message
        void notify(event::type type, const char* filename, unsigned int line, const std::string& message) const
        {
            // synchronize threads
            std::lock_guard<std::recursive_mutex> lk(this->m_lock);

            // do not allow recursive calls
            if (this->m_recurse_lock)
                return;

            // enable recursive locking
            this->m_recurse_lock = true;

            // create event
            logger::event event(std::chrono::system_clock::now(), type, format("%s:%d", filename, line), message);

            // filter message
            bool filter_succeed = false;

            try
            {
                filter_succeed = event.match(this->m_filter_types, this->m_filter_filenames, this->m_filter_messages);
            }
            catch (...) {}

            // only if filter succeeded
            if (filter_succeed)
            {
                // send to all loggers
                for (auto& p : this->m_loggers)
                {
                    try
                    {
                        if (p != nullptr)
                            p->notify(event);
                    }
                    catch (...) {}
                }
            }

            // duisable recursive locking
            this->m_recurse_lock = false;
        }

        // add logger
        void add(std::shared_ptr<base> p)
        {
            std::lock_guard<std::recursive_mutex> lk(this->m_lock);

            if(p != nullptr)
                this->m_loggers.push_back(p);
        }

        // add logger (templated version)
        template<class type, typename ... params> std::shared_ptr<base> add(params ... args)
        {
            auto p = std::make_shared<type>(args ...);

            add(p);

            return p;
        }

        // remove logger
        void remove(std::shared_ptr<base> p)
        {
            std::lock_guard<std::recursive_mutex> lk(this->m_lock);

            if (p != nullptr)
                this->m_loggers.remove(p);
        }

        // set filters
        void filter(std::list<event::type> types, std::list<std::regex> filenames, std::list<std::regex> messages)
        {
            std::lock_guard<std::recursive_mutex> lk(this->m_lock);

            this->m_filter_types = types;
            this->m_filter_filenames = filenames;
            this->m_filter_messages = messages;
        }

    private:

        // find window on initialization
        manager(void);

    private:
        mutable std::recursive_mutex m_lock;
        mutable bool m_recurse_lock;

        std::list<std::shared_ptr<base>> m_loggers;
        
        std::list<event::type> m_filter_types;
        std::list<std::regex> m_filter_filenames;
        std::list<std::regex> m_filter_messages;
    };

    /* auto_logger logger class */
    template<class type> class auto_logger
    {
    public:
        auto_logger(void)
        {
            this->m_pointer = std::make_shared<type>();

            get_instance<manager>()->add(this->m_pointer);
        }

        ~auto_logger(void)
        {
            get_instance<manager>()->remove(this->m_pointer);
        }

        std::shared_ptr<type> operator->(void)
        {
            return ptr();
        }

        std::shared_ptr<const type> operator->(void) const
        {
            return const_cast<auto_logger<type>*>(this)->ptr();
        }

        type& operator*(void)
        {
            return *ptr();
        }

        const type& operator*(void) const
        {
            return *(const_cast<auto_logger<type>*>(this)->ptr());
        }

    private:
        std::shared_ptr<type> ptr(void)
        {
            if (this->m_pointer == nullptr)
                throw_exception(null_pointer_exception);

            return this->m_pointer;
        }

    private:
        std::shared_ptr<type> m_pointer;
    };
};

using debug_logger_manager_t = debug::logger::manager;