#pragma once

#include "../debug/exception.h"
#include "../strings/fileparts.h"
#include "../debug/exdebug.h"
#include "../utils/singleton.h"
#include "../threads/lock.h"
#include "../storage/storage.h"
#include "../utils/base64.h"
#include "../threads/thread.h"
#include "../event/base.h"

#include "../exception.h"

#include <memory>
#include <unordered_map>
#include <mutex>
#include <string>
#include <filesystem>
#include <functional>
#include <list>

#ifdef _WIN32

#include <windows.h>

#else

#include <errno.h>
#include <sys/inotify.h>
#include <poll.h>
#include <unistd.h>

#endif

namespace resources
{

	/* file_change_monitor (win32 version) */
	class file_change_monitor
	{
	public:
		using callback_t = std::function<void(const std::string&)>;

	private:

		/* directory_watcher thread class */
		class directory_watcher : public thread::base
		{
		public:

			// start on ctor
			directory_watcher(file_change_monitor& parent, const std::string& directory) : m_parent(parent)
			{
				// save directory
				this->m_directory = directory;

				// initialize
#ifdef _WIN32
				this->m_monitor_handle = INVALID_HANDLE_VALUE;
#else
				this->m_monitor_handle = -1;
#endif

				// automaticaly start
				start();
			}

			// stop on ctor
			virtual ~directory_watcher(void)
			{
				try
				{
					stop();

					AUTOLOCK(this->m_lock);

					this->m_last_writes.clear();
				}
				catch (...) {}
			}

			// count watched resources
			size_t num_watched(void) const
			{
				AUTOLOCK(this->m_lock);

				return this->m_last_writes.size();
			}

			// check if file already exists
			bool has(const std::string& filename) const
			{
				AUTOLOCK(this->m_lock);

				return this->m_last_writes.find(filename) != this->m_last_writes.end();
			}

			// check if empty
			bool empty(void) const
			{
				AUTOLOCK(this->m_lock);

				return this->m_last_writes.empty();
			}

			// remove file
			bool remove(const std::string& filename)
			{
				AUTOLOCK(this->m_lock);

				auto it = this->m_last_writes.find(filename);

				if (it == this->m_last_writes.end())
					return false;

				this->m_last_writes.erase(it);

				return true;
			}

			// add file
			bool add(const std::string& filename)
			{
				AUTOLOCK(this->m_lock);

				// check first if file already exists
				if (this->m_last_writes.find(filename) != this->m_last_writes.end())
					return false;

				std::filesystem::file_time_type last_write;

				// can trigger error if file is removed
				try
				{
					last_write = std::filesystem::last_write_time(build_path({ this->m_directory, filename }));
				}
				catch (...) {}

				// add to list
				this->m_last_writes.emplace(std::make_pair(filename, last_write));

				// return successfuly
				return true;
			}

		private:

#ifdef _WIN32
			// a change has occured
			void process(void)
			{
				std::list<std::string> list;

				// populate list first (with lock)
				{
					AUTOLOCK(this->m_lock);

					// browse list
					for (auto& v : this->m_last_writes)
					{
						auto filename = build_path({ this->m_directory, v.first });

						std::filesystem::file_time_type last_write;
						
						// can trigger error if file is removed
						try
						{
							last_write = std::filesystem::last_write_time(filename);
						}
						catch (...) { }

						// skip if last write did not change
						if (v.second == last_write)
							continue;

						// otherelse add to list
						list.push_back(filename);

						// remember last write for later
						v.second = last_write;
					}
				}

				// notify change (without lock)
				for (auto& filename : list)
					trigger(filename);
			}
#endif

			// check for change
			virtual void run(void) override
			{
#ifdef _WIN32
				// copy handle
				HANDLE handle;

				{
					AUTOLOCK(this->m_lock);

					handle = this->m_monitor_handle;
				}

				if (handle == INVALID_HANDLE_VALUE)
					return;

				// wait for object
				DWORD dwStatus = WaitForSingleObject(handle, 100);

				if (dwStatus != WAIT_OBJECT_0)
					return;

				// check directory
				try
				{
					process();
				}
				catch (...) {}

				// process next change
				FindNextChangeNotification(handle);
#else
				// skip if no handle
				{
					AUTOLOCK(this->m_lock);

					if (this->m_monitor_handle == -1)
						return;
				}

				// poll handle
				{
					struct pollfd s;

					{
						AUTOLOCK(this->m_lock);

						s.fd = this->m_monitor_handle;
						s.events = POLL_IN;
					}

					// skip on timeout
					if (!poll(&s, 1, 100))
						return;

					// skip if no input data
					if ((s.revents & POLL_IN) == 0)
						return;
				}

				// we have data to read
				unsigned char buffer[4096] __attribute__((aligned(__alignof__(struct inotify_event))));

				while (true)
				{
					auto bytes_read = read(this->m_monitor_handle, buffer, sizeof(buffer));

					// nothing to read
					if (bytes_read <= 0)
						break;

					// scan all events
					size_t ofs = 0;

					while (ofs < bytes_read)
					{
						// get current and go to next
						auto p = (const struct inotify_event*)(&buffer[0] + ofs);
						ofs += sizeof(struct inotify_event) + p->len;

						// skip if no filename
						if (p->len == 0)
							continue;

						// check file
						std::string filename = build_path({ this->m_directory, p->name });

						{
							AUTOLOCK(this->m_lock);

							auto it = this->m_last_writes.find(p->name);

							// go to next if we are not watching that file
							if (it == this->m_last_writes.end())
								continue;

							std::filesystem::file_time_type last_write;

							// can trigger error if file is removed
							try
							{
								last_write = std::filesystem::last_write_time(filename);
							}
							catch (...) {}

							// skip if last write did not change
							if (it->second == last_write)
								continue;
						}

						// otherelse we have a match
						trigger(filename);
					}
				}

#endif
			}

			// trigger a file change
			void trigger(const std::string& filename)
			{
				try
				{
					this->m_parent.on_change(filename);
				}
				catch (exception::base& err)
				{
					_error("file change callback triggered exception \"%s\" !", err.to_string().c_str());
				}
				catch (std::exception& err)
				{
					_error("file change callback triggered exception \"%s\" !", err.what());
				}
				catch (...)
				{
					_error("file change callback triggered unknown exception !");
				}
			}

			// create handle on start
			virtual void on_start(void) override
			{
				AUTOLOCK(this->m_lock);

				// create file watcher object
#ifdef _WIN32
				this->m_monitor_handle = FindFirstChangeNotificationA(this->m_directory.c_str(), FALSE, FILE_NOTIFY_CHANGE_LAST_WRITE | FILE_NOTIFY_CHANGE_FILE_NAME);

				if (this->m_monitor_handle == INVALID_HANDLE_VALUE)
				{
					_warning("cannot create FindFirstChangeNotification object");
				}
#else
				this->m_monitor_handle = inotify_init1(IN_NONBLOCK);

				if (this->m_monitor_handle == -1)
				{
					_warning("cannot create inotify object");
				}

				// add watcher
				auto id = inotify_add_watch(this->m_monitor_handle, this->m_directory.c_str(), IN_CREATE | IN_DELETE | IN_MOVED_FROM | IN_MODIFY | IN_ATTRIB);

				if (id == -1)
				{
					_warning("cannot create file watch for %s (%s)", this->m_directory.c_str(), strerror(errno));
				}
#endif
			}

			// close handle on stop
			virtual void on_stop(void) override
			{
				AUTOLOCK(this->m_lock);

				// close monitor object
#ifdef _WIN32
				if (this->m_monitor_handle != INVALID_HANDLE_VALUE)
					FindCloseChangeNotification(this->m_monitor_handle);

				this->m_monitor_handle = INVALID_HANDLE_VALUE;
#else
				if (this->m_monitor_handle != -1)
					close(this->m_monitor_handle);

				this->m_monitor_handle = -1;
#endif
			}

		private:
			mutable std::mutex m_lock;

			file_change_monitor& m_parent;

			std::unordered_map<std::string, std::filesystem::file_time_type> m_last_writes;
			std::string m_directory;

#ifdef _WIN32
			HANDLE m_monitor_handle;
#else
			int m_monitor_handle;
#endif
		};

	public:

		// ctor
		file_change_monitor(callback_t callback)
		{
			this->m_callback = callback;
		}

		// clear on dtor
		~file_change_monitor(void)
		{
			try
			{
				AUTOLOCK(this->m_lock);

				for (auto& v : this->m_directories)
					if (v.second != nullptr)
						v.second->stop();

				this->m_directories.clear();
			}
			catch (...) {}
		}

		// count files being watched
		size_t num_watched(void) const
		{
			size_t num = 0;

			AUTOLOCK(this->m_lock);

			for (auto& v : this->m_directories)
				if (v.second != nullptr)
					num += v.second->num_watched();

			return num;
		}

		// add file to list
		bool add(const std::string& filename)
		{
			_debug("monitoring file %s", filename.c_str());

			AUTOLOCK(this->m_lock);

			auto s = split_file_path(filename);

			// check if directory is already listed
			auto it = this->m_directories.find(s.directory);

			if (it != this->m_directories.end())
			{
				if (__ISNULLPTR(it->second))
					return false;

				return it->second->add(s.file);
			}

			// otherelse create one
			auto p = std::make_shared<directory_watcher>(*this, s.directory);

			if(!p->add(s.file))
				return false;

			this->m_directories.emplace(std::make_pair(s.directory, p));

			// return successful
			return true;
		}

		// remove file
		bool remove(const std::string& filename)
		{
			_debug("remove monitoring on file %s", filename.c_str());

			AUTOLOCK(this->m_lock);

			auto s = split_file_path(filename);

			// find directory
			auto it = this->m_directories.find(s.directory);

			if (it == this->m_directories.end())
			{
				_warning("directory %s not found!", s.directory.c_str());

				return false;
			}

			// should not happen
			if (__ISNULLPTR(it->second))
				return false;

			// remove file
			if (!it->second->remove(s.file))
			{
				_warning("cannot remove file %s!", s.file.c_str());
				
				return false;
			}

			// remove directory watcher if empty
			if (it->second->empty())
				this->m_directories.erase(it);

			// return successful
			return true;
		}

	private:

		// call callback function on change
		void on_change(const std::string& filename)
		{
			if(this->m_callback)
				this->m_callback(filename);
		}

	private:
		mutable std::mutex m_lock;
		std::unordered_map<std::string, std::shared_ptr<directory_watcher>> m_directories;
		callback_t m_callback;
	};
};