#pragma once

#include "exception.h"
#include "manager.h"

#include <memory>
#include <string>

namespace resources
{
	/* sync resources class */
	template<class type, bool async = false> class proxy
	{
	public:

		// ctor
		proxy(void) {}

		// ctor by name
		proxy(const std::string& name)
		{
			load(name);
		}

		// reset on dtor
		~proxy(void)
		{
			try
			{
				reset();
			}
			catch (...) {}
		}

		// load
		void load(const std::string& name)
		{
			reset();

			this->m_cache_pointer = get_cache_pointer<type>(name, async);
		}

		// cleanup on reset
		void reset(void)
		{
			// skip if already null
			if (this->m_cache_pointer == nullptr)
				return;

			// always wait on reset/dtor to avoid having something in the loader queue that blocks the file monitoring process
			wait();

			// copy name
			auto name = this->m_cache_pointer->name();

			// reset first
			this->m_cache_pointer.reset();

			// cleanup cache pointer
			cleanup_cache_pointer<type>(name);
		}

		// wait for loading
		bool wait(const double timeout = 0.0) const
		{
			if (this->m_cache_pointer == nullptr)
				throw_exception(resource_uninitialized_exception);

			return this->m_cache_pointer->wait(timeout);
		}

		// check if pending
		bool is_pending(void) const
		{
			if (this->m_cache_pointer == nullptr)
				return false;

			return this->m_cache_pointer->is_pending();
		}

		// retrieve name
		const std::string& name(void) const
		{
			if (this->m_cache_pointer == nullptr)
				throw_exception(resource_uninitialized_exception);

			return this->m_cache_pointer->name();
		}

		// force reload
		void reload(void)
		{
			if (this->m_cache_pointer == nullptr)
				throw_exception(resource_uninitialized_exception);

			this->m_cache_pointer->reload(async);
		}

		// return pointer
		std::shared_ptr<const type> operator->(void) const
		{
			return ptr();
		}

		// return reference
		const type& operator*(void) const
		{
			return *ptr();
		}

		// get raw pointer
		std::shared_ptr<const type> get(void) const
		{
			if (this->m_cache_pointer == nullptr)
				throw_exception(resource_uninitialized_exception);

			return this->m_cache_pointer->get();
		}

	protected:

		// check and return pointer
		std::shared_ptr<const type> ptr(void) const
		{
			if (this->m_cache_pointer == nullptr)
				throw_exception(resource_uninitialized_exception);

			// load
			auto p = get();

			if (p == nullptr)
				throw_exception(null_pointer_exception);

			return p;
		}

	private:
		std::shared_ptr<const manager::cache_t<type>> m_cache_pointer;
	};
};